"""Source class"""

import datetime as dt
import warnings
from typing import Optional

import cryptocompare
import pandas as pd
from cryptocmd import CmcScraper

START = dt.datetime(2009, 1, 1)


class Source:

    def __init__(self, source: str, api_key: Optional[str] = None):
        self.source = source
        self.api_key = api_key

    def get_data(self, coin: str, fiat: str) -> pd.DataFrame:
        """Get historical OHLC data

        entrypoint to child classes method

        Args:
            coin (str): coin symbol
            fiat (str): currency symbol

        Returns:
            DataFrame: historical OHLC data

        Raises:
            ValueError: source not available
        """
        if self.source == "cryptocompare":
            try:
                cryptocompare.cryptocompare._set_api_key_parameter(self.api_key)
                data = cryptocompare.get_historical_price_day_from(
                    coin=coin, currency=fiat, fromTs=START
                )
                data = pd.DataFrame(data)
                data["time"] = pd.to_datetime(data["time"], unit="s")
                data.rename({"time": "Date", "close": "Close"}, axis=1, inplace=True)
                return data[["Date", "Close"]]
            except Exception as e:
                raise ValueError(f"Error getting data from 'cryptocompare' source: {e}")

        elif self.source == "coinmarketcap-free":
            warnings.warn(
                "The 'coinmarketcap-free' source is broken,"
                " and if a fix is not found,"
                " it will be removed in the future.",
            )
            try:
                scraper = CmcScraper(coin, fiat=fiat)
                scraper.get_data()
                return scraper.get_dataframe()[["Date", "Close"]]
            except Exception as e:
                raise ValueError(
                    f"Error getting data from 'coinmarketcap-free' source: {e}"
                )

        elif self.source == "coinmarketcap":
            raise NotImplementedError("CoinMarketCap API v2 not implemented")

        else:
            raise ValueError(f"source '{self.source}' not available")



"""common utils for artists"""

from __future__ import annotations

import datetime as dt
from typing import TYPE_CHECKING

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import pandas as pd

if TYPE_CHECKING:
    from ..bitcoin import Bitcoin


class ColorBar:
    """color bar

    Args:
        bitcoin (Bitcoin): bitcoin object

    Attributes:
        norm (matplotlib.Normalize): normalize
        cmap (matplotlib.Colormap): colormap
    """

    def __init__(self, bitcoin: Bitcoin):
        self._set_cmap(bitcoin)

    def _set_cmap(self, bitcoin: Bitcoin) -> None:
        """set colors"""
        self.norm = mcolors.Normalize(
            vmin=bitcoin.prices["distance_ath_perc"].min(),
            vmax=bitcoin.prices["distance_ath_perc"].max(),
        )
        self.cmap = plt.get_cmap("cool")


class ProgressLabels:
    """progress labels

    Args:
        bitcoin (Bitcoin): bitcoin object

    Attributes:
        labels (Series): labels
        predicted_halving_str (str): predicted halving string
    """

    def __init__(self, bitcoin: Bitcoin):
        self._create_labels(bitcoin)

    def _create_labels(self, bitcoin: Bitcoin) -> None:
        """create labels for [0, 0.25, 0.50, 0.75] percent of cycle progress"""
        self.labels = []
        self._get_moments(bitcoin)

    def _get_moments(self, bitcoin) -> None:
        """get moments for each progress value"""
        # for each progress value, get the corresponding dates
        # (for each cycle_id get the first date matching the progress)
        for progress in [0.00, 0.25, 0.50, 0.75]:
            self.labels.append(
                bitcoin.prices[
                    abs((bitcoin.prices["cycle_progress"] - progress)) < 0.0005
                ]
                .groupby("cycle_id")
                .first()
                .reset_index()
            )
        self.labels = pd.concat(self.labels)

        # round cycle progress to 2 decimal places
        self.labels["cycle_progress"] = self.labels["cycle_progress"].apply(
            lambda x: round(x, 2)
        )
        # # concat groupby objects as formatted strings
        self.labels = self.labels.groupby("cycle_progress")["Date"].apply(
            lambda x: "".join(
                f"{label}\n" for label in x.dt.strftime("%d-%m-%Y").to_list()
            )
        )

        self._add_predicted(bitcoin)

    def _add_predicted(self, bitcoin) -> None:
        """adds predicted halving date to labels"""
        # get current cycle length
        expected_length = bitcoin.halvings["cycle_length"].dropna().iloc[-1]
        # list of predicted dates (0.25, 0.5 and 0.75 of the cycle)
        predicted_dates = [bitcoin.predicted_halving_date] + [
            bitcoin.halvings["Date"].dropna().iloc[-2] + dt.timedelta(days=x)
            for x in [x * expected_length for x in [0.25, 0.5, 0.75]]
        ]
        # string and add to labels
        for i, date in enumerate(predicted_dates):
            predicted_string = r"$\bf{{{}}}$".format(date.strftime("%d-%m-%Y"))
            self.labels.iloc[i] = self.labels.iloc[i] + predicted_string



"""artist module"""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Literal, Union

from .static import StaticArtist

if TYPE_CHECKING:
    from ..bitcoin import Bitcoin

THEMES = {
    "light": {
        "background": "white",
        "text": "black",
        "grid": "darkgrey",
        "now_line": "darkgrey",
        "halving_line": "lightgreen",
        "ath_marker": "black",
    },
    "dark": {
        "background": "black",
        "text": "white",
        "grid": "lightgrey",
        "now_line": "lightgrey",
        "halving_line": "lightgreen",
        "ath_marker": "white",
    },
}


class Artist:
    """Artist for plotting"""

    def __init__(
        self,
        bitcoin: Bitcoin,
        kind: Literal["static", "dynamic"],
        theme: Union[Literal["light", "dark"], dict],
    ):
        """Artist

        Artist for plotting, either static or dynamic.
        Static artist uses `matplotlib` and dynamic artist uses `plotly`.

        Arguments:
            bitcoin (Bitcoin): bitcoin object.
            kind (Literal["static", "dynamic"], optional): type of artist.
                Defaults to "static".
            theme (Union[Literal["light", "dark"], dict], optional): theme colors.
                Defaults to "light".


        Attributes:
            artist (StaticArtist): artist for plotting
            theme (dict): theme colors

        Raises:
            NotImplementedError: _description_
            ValueError: _description_
        """
        self.artist = None
        self.theme = None

        # unwrap theme
        self.__unwrap_theme(theme)

        # set artist
        if kind == "static":
            self._kind = kind
            self.artist = StaticArtist(bitcoin, theme=self.theme)
        elif kind == "dynamic":
            # TODO: implement dynamic artist (plotly)
            raise NotImplementedError
        else:
            raise ValueError("kind must be 'static' or 'dynamic'")

    def __unwrap_theme(
        self,
        theme: Union[Literal["light", "dark"], dict],
    ) -> None:
        """unwrap theme

        Args:
            theme (Union[Literal["light", "dark"], dict], optional): theme colors.
        """
        if isinstance(theme, str) and theme in THEMES:
            self.theme = THEMES[theme]
        # if dict, check if it has the right keys
        elif isinstance(theme, dict):
            # if theme is a dictionary with the same keys as the default themes
            # then use it
            if set(theme.keys()) == set(THEMES["light"].keys()):
                self.theme = theme
            else:
                # else use the light theme and update the keys
                for key in theme.keys():
                    self.theme = THEMES["light"]
                    if key not in THEMES["light"]:
                        # warning message
                        warnings.warn(f"Theme key '{key}' not recognized. Ignoring it.")
                    else:
                        self.theme[key] = theme[key]
        else:
            raise ValueError(
                "theme must be 'light', 'dark' or a dictionary with the theme colors"
            )

    @property
    def kind(self) -> str:
        return self._kind

    def plot(self, **kwargs):
        """wrapper for plot method of artist

        Args:
            \\*\\*kwargs: keyword arguments to plotting method
        """
        return self.artist.plot(**kwargs)



"""CoinmarketCap API wrapper."""

import requests

BASE_URL = "https://pro-api.coinmarketcap.com/v2"
ENDPOINT = "/cryptocurrency/ohlcv/historical"


class CoinMarketCap:
    """CoinMarketCap v2 API wrapper"""

    def __init__(self, api_key: str):
        self.api_key = api_key

    def get_historical_data(self, symbol: str, start: str, end: str) -> dict:
        """Get historical OHLC data from CoinMarketCap

        Args:
            symbol (str): cryptocurrency symbol
            start (str): start date in format "YYYY-MM-DD"
            end (str): end date in format "YYYY-MM-DD"

        Returns:
            dict: historical OHLC data
        """
        headers = {
            "Accepts": "application/json",
            "X-CMC_PRO_API_KEY": self.api_key,
        }
        params = {
            "symbol": symbol,
            "time_start": start,
            "time_end": end,
        }
        url = f"{BASE_URL}{ENDPOINT}"
        response = requests.get(url, headers=headers, params=params)
        return response.json()



"""bitcoin module"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Literal, Union

from ..artist import Artist
from .halvings import Halvings, get_halving_data
from .prices import Prices

# TODO: for now its assuming that the
#  cycle length is the same as previous

if TYPE_CHECKING:
    import matplotlib.figure


class Bitcoin:
    """Bitcoin

    Bitcoin class stores bitcoin prices and halvings data.
    All sources require an API key.

    Available data sources:
    - coinmarketcap
    - cryptocompare
    - coinmarketcap-free (**legacy broken source**)

    Plot bitcoin prices and halvings data.

    Args:
        source (str, optional): data source. Defaults to "cryptocompare".
        currency (str, optional): currency. Defaults to "USD".
        api_key (str, optional): API key. Defaults to None.

    Attributes:
        prices (DataFrame): bitcoin prices
        halvings (DataFrame): bitcoin halvings
        predicted_halving_date (datetime): predicted halving date
        predicted_halving_block (int): predicted halving block
    """

    def __init__(
        self,
        source: str = "coinmarketcap-free",
        currency: str = "USD",
        api_key: str = None,
    ):
        # get processed price data
        self.prices = Prices(currency=currency, source=source, api_key=api_key).data
        # save halving data
        self.halvings = Halvings().data
        # predicted next halving date and block
        self.predicted_halving_date, self.predicted_halving_block = get_halving_data()

    def plot(
        self,
        kind: str = "static",
        from_date: Union[str, datetime.datetime] = None,
        theme: Union[Literal["light", "dark"], dict] = "light",
        **plotting_kwargs,
    ) -> matplotlib.figure.Figure:
        """plot

        Args:
            kind (str, optional): plot kind. Defaults to "static".
            from_date (Union[str, datetime.datetime], optional): start date.
                Defaults to None, which fetches all data.
            \\*\\*plotting_kwargs: additional keyword arguments to Artist's
                plotting method.
            theme (Union[Literal["light", "dark"],dict], optional): theme
                for the plot. Defaults to "light". If a dictionary is passed,
                it should contain one of following keys. It's not required to
                pass all of them, only the ones you want to change from the
                default `light` theme:
                - background: background color
                - text: text color
                - grid: grid color
                - now_line: now line color
                - halving_line: halving line color
                - ath_marker: all-time high marker color

        Returns:
            matplotlib.figure.Figure: figure object
        """
        # update plotting kwargs
        plotting_kwargs.update({"from_date": from_date})
        # plot
        return Artist(bitcoin=self, kind=kind, theme=theme).plot(**plotting_kwargs)



"""prices module"""

from typing import Optional

import pandas as pd

from .halvings import Halvings
from .sources import Source


def _find_ath(dataframe: pd.DataFrame) -> pd.DataFrame:
    """find all-time high

    Args:
        dataframe (DataFrame): historical OHLC data with
            "Close" column

    Returns:
        DataFrame: historical OHLC data with
            "ATH" and "distance_ath_perc" columns
    """
    # find ATH
    dataframe["ATH"] = dataframe["Close"].cummax()
    # find distance from ATH in percentage
    dataframe["distance_ath_perc"] = (
        dataframe["Close"] - dataframe["ATH"]
    ) / dataframe["ATH"]
    return dataframe


def _find_cycle_progress(dataframe: pd.DataFrame) -> pd.DataFrame:
    """find cycle progress

    Args:
        dataframe (DataFrame): historical OHLC data with
            "cycle" and "cycle_length" columns

    Returns:
        DataFrame: historical OHLC data with
            "cycle_progress" column
    """
    dataframe["cycle_progress"] = (
        (dataframe["Date"] - dataframe["Halving"]).dt.days
    ) / dataframe["cycle_length"]
    return dataframe


class Prices:
    """Get historical OHLC data and set metrics

    Prices class to get historical OHLC data and halving data,
    and process it.

    Gets historical OHLC data using desired `source`.

    Sets the following metrics on the OHLC dataframe:
    - ATH
    - distance from ATH in percentage
    - cycle progress

    Args:
        source (str): source to get historical OHLC data
        api_key (str): API key for source

    Attributes:
        coin (str): coin symbol
        source(str): source to get historical OHLC data
        fiat (str): currency symbol
        data (DataFrame): historical OHLC data
        halvings (DataFrame): halving data
    """

    coin: str = "BTC"
    source: Optional[str] = None
    fiat: Optional[str] = None
    data: Optional[pd.DataFrame] = None
    halvings: Optional[pd.DataFrame] = None

    def __init__(self, currency: str, source: str, api_key: str):
        self.source = source
        self.fiat = currency
        # get price data
        self.data = Source(source, api_key).get_data(self.coin, self.fiat)
        # get halving data
        self.halvings = Halvings().data
        # format DataFrame
        self._fmt_df()
        # set metrics
        self._set_metrics()

    def _fmt_df(self) -> None:
        """Format DataFrame"""
        # make Date UTC aware
        self.data["Date"] = pd.to_datetime(self.data["Date"]).dt.tz_localize("UTC")
        # add column to merged df so to keep track of halving dates
        halvings = self.halvings.copy()
        halvings["Halving"] = halvings["Date"]
        self.data = self.data.merge(
            halvings,
            how="outer",  # keep all halving dates
            on="Date",
        )
        # # fill NaNs due to merge only on subset of columns
        self.data[["block", "cycle_length", "cycle_id", "Halving", "reward"]] = (
            self.data[["block", "cycle_length", "cycle_id", "Halving", "reward"]]
            .infer_objects()
            .ffill()
        )
        # remove Halvings outside of prices datarange
        self.data = self.data[self.data["Close"].notna()]
        # sort by date (ATH calculation requires ascending order)
        self.data = self.data.sort_values("Date").reset_index(drop=True)

    def _set_metrics(self) -> None:
        """set metrics to prices DataFrame"""
        # find ATH
        self.data = _find_ath(self.data)
        # find cycle progress
        self.data = _find_cycle_progress(self.data)



"""static artist module"""

from __future__ import annotations

import copy
import datetime
from importlib.metadata import version
from typing import TYPE_CHECKING, Union

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from .utils import ColorBar, ProgressLabels

if TYPE_CHECKING:
    from ..bitcoin import Bitcoin


class StaticArtist:
    """static artist

    Using `matpotlib` to plot statically.

    Args:
        bitcoin (Bitcoin): bitcoin object
        colorbar (ColorBar): color bar object
        theme (dict): a dictionary with the theme colors

    Attributes:
        bitcoin (Bitcoin): bitcoin object
        colorbar (ColorBar): color bar object
        theme (dict): a dictionary with the theme colors
    """

    def __init__(self, bitcoin: Bitcoin, theme: dict = "light"):
        self.bitcoin = copy.copy(bitcoin)
        self.colorbar = ColorBar(self.bitcoin)
        self.theme = theme

        # move to a colorbar method
        self._set_colors()

    def _set_colors(self):
        """set colors

        Create a new column in the DataFrame with the color
        """
        # TODO: this is a colorbar method
        self.bitcoin.prices["color"] = self.bitcoin.prices["distance_ath_perc"].apply(
            lambda x: mcolors.to_hex(self.colorbar.cmap(self.colorbar.norm(x)))
        )

    def plot(self, from_date: Union[str, datetime.datetime]):
        """plot

        Args:
            from_date (Union[str, datetime.datetime]): start date
        """
        # Create a polar subplot
        self.f, self.axes = plt.subplots(
            1, 1, subplot_kw=dict(polar=True), figsize=(10, 10)
        )
        # Set background color
        self.f.set_facecolor(self.theme["background"])
        self.axes.set_facecolor(self.theme["background"])

        # Plot data
        self.add_data(from_date=from_date)

        # now point
        self.add_now()

        # halving
        self.add_halving()

        # # Plot ATHs
        self.add_aths()

        # # # plot bottoms
        # self.add_bottoms()

        # format graph
        self.format_chart()

        # legend
        self.add_legend()

        # colorbar
        self.add_colorbar()

        # image creation date and copyright
        self.add_watermark()

        # necessary althought the warning
        # otherwise a savefig on the fig would cut
        self.f.tight_layout()

        # return figure
        return self.f

    def add_watermark(self) -> None:
        """add watermark to plot"""
        try:
            # python ^3.10
            date = datetime.datetime.now(datetime.UTC)
        except AttributeError:
            date = datetime.datetime.utcnow()
        date_text = date.strftime("%Y-%m-%d %H:%M UTC")
        self.axes.annotate(
            (f"{date_text}\n" f"© giocaizzi/btc-cycles : {version('btc-cycles')}"),
            xy=(0.5, 0.33),
            xycoords="axes fraction",
            textcoords="axes fraction",
            fontsize=8,
            ha="center",
            va="center",
            color="darkgrey",
        )

    def add_data(self, from_date: Union[str, datetime.datetime]) -> None:
        """add data to plot

        This method filters the data inplace.

        Args:
            from_date (Union[str, datetime.datetime]): start date
        """
        if from_date is not None:
            # filter data
            self.bitcoin.prices = self.bitcoin.prices[
                self.bitcoin.prices.Date >= from_date
            ]

        # plot data
        self.axes.scatter(
            self.bitcoin.prices["cycle_progress"] * 2 * np.pi,
            self.bitcoin.prices["Close"].to_numpy(),
            s=3,
            # markersize=1,
            c=self.bitcoin.prices["color"],
            zorder=9,
        )

    def format_chart(self) -> None:
        """format axes"""
        # Set y-axis to logarithmic scale
        self.axes.set_rscale("log")

        # Set direction (1 for clockwise, -1 for counterclockwise)
        self.axes.set_theta_direction(-1)
        self.axes.set_theta_offset(np.pi / 2.0)

        grid_intervals = [
            0.001,
            0.01,
            0.1,
            1,
            10,
            100,
            1000,
            10000,
            100000,
            1000000,
        ]

        labels = [
            "0.001",
            "0.01",
            "0.1",
            "1",
            "10",
            "100",
            "1k",
            "10k",
            "100k",
            "1M",
        ]

        start_index = next(
            i
            for i, v in enumerate(grid_intervals)
            if v >= self.bitcoin.prices.Close.min()
        )

        # set gridline color
        self.axes.grid(color=self.theme["grid"])

        # Set r gridlines
        self.axes.set_rgrids(
            grid_intervals[start_index:],
            labels=labels[start_index:],
        )

        # Set xticks and labels
        self.axes.set_xticks(
            np.linspace(0, 2 * np.pi, 4, endpoint=False),
        )
        self.axes.set_xticklabels(
            ProgressLabels(self.bitcoin).labels,
            fontsize=8,
        )

        # r label
        self.axes.set_rlabel_position(0)

        # y label
        self.axes.set_ylabel("Price (USD)", rotation=0)
        self.axes.yaxis.set_label_coords(0.5, 1.01)

        # ticks params
        self.axes.tick_params(
            axis="both", which="major", pad=30, colors=self.theme["text"]
        )

        # edge color
        [spine.set_edgecolor("lightgrey") for spine in self.axes.spines.values()]

    def add_bottoms(self) -> None:
        """add bottoms to plot"""
        self.axes.scatter(
            self.bitcoin.prices["cycle_progress"] * 2 * np.pi,
            self.bitcoin.prices["Close"],
            marker="o",
            c="r",
            s=20,
            zorder=10,
        )

    def add_aths(self) -> None:
        """add all time highs to plot"""
        aths = self.bitcoin.prices[self.bitcoin.prices["distance_ath_perc"] == 0]
        self.axes.scatter(
            aths["cycle_progress"] * 2 * np.pi,
            aths["Close"],
            marker="x",
            c=self.theme["ath_marker"],
            s=20,
            zorder=10,
        )

    def add_halving(self) -> None:
        """add halving to plot"""
        self.axes.vlines(
            0,
            self.bitcoin.prices["Close"].min(),
            1000000,
            color=self.theme["halving_line"],
            linewidth=3,
            zorder=0,
        )

    def add_now(self) -> None:
        self.axes.scatter(
            self.bitcoin.prices["cycle_progress"].to_numpy()[-1] * 2 * np.pi,
            self.bitcoin.prices["Close"].to_numpy()[-1],
            marker="D",
            c=self.bitcoin.prices["color"].to_numpy()[-1],
            s=50,
            zorder=8,
        )
        # now line
        self.axes.vlines(
            self.bitcoin.prices["cycle_progress"].to_numpy()[-1] * 2 * np.pi,
            self.bitcoin.prices["Close"].min(),
            self.bitcoin.prices["Close"].to_numpy()[-1],
            color=self.theme["now_line"],
            linestyle="--",
            zorder=8,
        )

    def add_legend(self) -> None:
        """add legend and title to plot"""
        legend = self.axes.legend(
            [
                "BTC/USD",
                "Today BTC/USD Close",
                "Today",
                "Halving day",
                "All time high (ATH)",
            ],
            loc="upper left",
            bbox_to_anchor=(-0.075, 1.05),
            fontsize=10,
            # title="$\\bf{Legend}$",
            title_fontsize="13",
            # labelspacing=1.5,
            frameon=False,
        )
        # Set the color of legend text
        for text in legend.get_texts():
            text.set_color(self.theme["text"])

        # title
        title = self.axes.set_title(
            "Bitcoin price halving cycles",
            fontdict={"fontsize": 15, "fontweight": "bold"},
            pad=20,
        )
        title.set_color(self.theme["text"])

    def add_colorbar(self) -> None:
        """add colorbar to plot"""
        sm = plt.cm.ScalarMappable(cmap=self.colorbar.cmap, norm=self.colorbar.norm)
        inset_ax = inset_axes(
            self.axes,
            width="2.5%",
            height="25%",
            # loc="upper right",
            bbox_to_anchor=(0, 0, 0.95, 1.05),
            bbox_transform=self.axes.transAxes,
            borderpad=0,
        )
        cbar = self.f.colorbar(
            sm,
            cax=inset_ax,
            orientation="vertical",
            ticks=[0, -0.2, -0.4, -0.6, -0.8, -1],
        )
        cbar.ax.set_yticklabels(
            ["ATH", "-20%", "-40%", "-60%", "-80%", "-100%"], color=self.theme["text"]
        )
        inset_ax.set_title(
            "Distance from ATH",
            fontdict={"fontweight": "bold", "color": self.theme["text"]},
            pad=10,
        )



"""halving module"""

import datetime
import json
from pathlib import Path

import pandas as pd
import requests

URL = "https://api.watcher.guru/bitcoinhalving/predictions"


def get_halving_data() -> tuple:
    """Get next halving data

    Gets the next halving data (date, blocknumber)
    from the watcher.guru API.

    Returns:
        tuple: date (datetime), block (int)
    """
    response = requests.get(URL).json()
    date = datetime.datetime.fromtimestamp(response["target"]["predicted_timestamp"])
    # localize utc
    date = date.replace(tzinfo=datetime.timezone.utc)
    block = int(response["target"]["block_number"])
    return date, block


def update_predicted_halving_date(data: pd.DataFrame) -> pd.DataFrame:
    """Update predicted halving date

    Update the predicted halving date in the halving dataframe
    by matching the block number.

    Args:
        data (DataFrame): halving data

    Returns:
        DataFrame: updated halving data
    """
    # TODO: what happends if response is not
    # updated in time? (like in 2024)
    date, block = get_halving_data()
    # update predicted halving date
    data.loc[data["block"] == block, "Date"] = date
    return data


class Halvings:
    """Halvings

    Halvings class that contains halving cycles data.

    Attributes:
        data (pd.DataFrame): halving data
    """

    def __init__(self):
        halvings_path = Path(__file__).resolve().parent / "halvings.json"
        # load json data into DataFrame
        with open(halvings_path, "r") as f:
            self.data = pd.DataFrame(json.load(f)).T
        # convert date to datetime
        self.data["date"] = pd.to_datetime(self.data["date"])
        self.data.index.name = "block"
        # Date with capital D
        self.data.rename(columns={"date": "Date"}, inplace=True)
        # reset index
        self.data.reset_index(inplace=True)
        self.data["block"] = self.data["block"].astype(int)
        # update predicted halving date
        self.data = update_predicted_halving_date(self.data)
        # cycle length, forced to int
        self.data["cycle_length"] = (self.data["Date"].diff().dt.days).shift(-1)
        # cycle id
        self.data["cycle_id"] = range(1, len(self.data) + 1)
